@@ -12,14 +12,14 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import contextlib, os.path, re, tempfile, shlex
+import abc, contextlib, enum, os.path, re, tempfile, shlex
 import subprocess
 
 from ..linkers import StaticLinker
 from .. import coredata
 from .. import mlog
 from .. import mesonlib
-from ..mesonlib import EnvironmentException, MesonException, version_compare, Popen_safe
+from ..mesonlib import EnvironmentException, MesonException, OrderedSet, version_compare, Popen_safe
 
 """This file contains the data files of all compilers Meson knows
 about. To support a new compiler, add its information below.
@@ -124,31 +124,35 @@ def is_library(fname):
     return suffix in lib_suffixes
 
 gnulike_buildtype_args = {'plain': [],
-                          # -O0 is passed for improved debugging information with gcc
-                          # See https://github.com/mesonbuild/meson/pull/509
-                          'debug': ['-O0', '-g'],
-                          'debugoptimized': ['-O2', '-g'],
-                          'release': ['-O3'],
-                          'minsize': ['-Os', '-g']}
+                          'debug': [],
+                          'debugoptimized': [],
+                          'release': [],
+                          'minsize': [],
+                          'custom': [],
+                          }
 
 armclang_buildtype_args = {'plain': [],
                            'debug': ['-O0', '-g'],
                            'debugoptimized': ['-O1', '-g'],
                            'release': ['-Os'],
-                           'minsize': ['-Oz']}
+                           'minsize': ['-Oz'],
+                           'custom': [],
+                           }
 
 arm_buildtype_args = {'plain': [],
                       'debug': ['-O0', '--debug'],
                       'debugoptimized': ['-O1', '--debug'],
                       'release': ['-O3', '-Otime'],
                       'minsize': ['-O3', '-Ospace'],
+                      'custom': [],
                       }
 
 msvc_buildtype_args = {'plain': [],
-                       'debug': ["/MDd", "/ZI", "/Ob0", "/Od", "/RTC1"],
-                       'debugoptimized': ["/MD", "/Zi", "/O2", "/Ob1"],
-                       'release': ["/MD", "/O2", "/Ob2"],
-                       'minsize': ["/MD", "/Zi", "/Os", "/Ob1"],
+                       'debug': ["/ZI", "/Ob0", "/Od", "/RTC1"],
+                       'debugoptimized': ["/Zi", "/Ob1"],
+                       'release': ["/Ob2", "/Gw"],
+                       'minsize': ["/Zi", "/Gw"],
+                       'custom': [],
                        }
 
 apple_buildtype_linker_args = {'plain': [],
@@ -156,6 +160,7 @@ apple_buildtype_linker_args = {'plain': [],
                                'debugoptimized': [],
                                'release': [],
                                'minsize': [],
+                               'custom': [],
                                }
 
 gnulike_buildtype_linker_args = {'plain': [],
@@ -163,6 +168,7 @@ gnulike_buildtype_linker_args = {'plain': [],
                                  'debugoptimized': [],
                                  'release': ['-Wl,-O1'],
                                  'minsize': [],
+                                 'custom': [],
                                  }
 
 arm_buildtype_linker_args = {'plain': [],
@@ -170,6 +176,7 @@ arm_buildtype_linker_args = {'plain': [],
                              'debugoptimized': [],
                              'release': [],
                              'minsize': [],
+                             'custom': [],
                              }
 
 msvc_buildtype_linker_args = {'plain': [],
@@ -180,6 +187,7 @@ msvc_buildtype_linker_args = {'plain': [],
                               # REF implies ICF.
                               'release': ['/OPT:REF'],
                               'minsize': ['/INCREMENTAL:NO', '/OPT:REF'],
+                              'custom': [],
                               }
 
 java_buildtype_args = {'plain': [],
@@ -187,48 +195,55 @@ java_buildtype_args = {'plain': [],
                        'debugoptimized': ['-g'],
                        'release': [],
                        'minsize': [],
+                       'custom': [],
                        }
 
 rust_buildtype_args = {'plain': [],
-                       'debug': ['-C', 'debuginfo=2'],
-                       'debugoptimized': ['-C', 'debuginfo=2', '-C', 'opt-level=2'],
-                       'release': ['-C', 'opt-level=3'],
-                       'minsize': [], # In a future release: ['-C', 'opt-level=s'],
+                       'debug': [],
+                       'debugoptimized': [],
+                       'release': [],
+                       'minsize': [],
+                       'custom': [],
                        }
 
 d_gdc_buildtype_args = {'plain': [],
-                        'debug': ['-g', '-O0'],
-                        'debugoptimized': ['-g', '-O'],
-                        'release': ['-O3', '-frelease'],
+                        'debug': [],
+                        'debugoptimized': ['-finline-functions'],
+                        'release': ['-frelease', '-finline-functions'],
                         'minsize': [],
+                        'custom': [],
                         }
 
 d_ldc_buildtype_args = {'plain': [],
-                        'debug': ['-g', '-O0'],
-                        'debugoptimized': ['-g', '-O'],
-                        'release': ['-O3', '-release'],
+                        'debug': [],
+                        'debugoptimized': ['-enable-inlining', '-Hkeep-all-bodies'],
+                        'release': ['-release', '-enable-inlining', '-Hkeep-all-bodies'],
                         'minsize': [],
+                        'custom': [],
                         }
 
 d_dmd_buildtype_args = {'plain': [],
-                        'debug': ['-g'],
-                        'debugoptimized': ['-g', '-O'],
-                        'release': ['-O', '-release'],
+                        'debug': [],
+                        'debugoptimized': ['-inline'],
+                        'release': ['-release', '-inline'],
                         'minsize': [],
+                        'custom': [],
                         }
 
 mono_buildtype_args = {'plain': [],
-                       'debug': ['-debug'],
-                       'debugoptimized': ['-debug', '-optimize+'],
+                       'debug': [],
+                       'debugoptimized': ['-optimize+'],
                        'release': ['-optimize+'],
                        'minsize': [],
+                       'custom': [],
                        }
 
 swift_buildtype_args = {'plain': [],
-                        'debug': ['-g'],
-                        'debugoptimized': ['-g', '-O'],
-                        'release': ['-O'],
+                        'debug': [],
+                        'debugoptimized': [],
+                        'release': [],
                         'minsize': [],
+                        'custom': [],
                         }
 
 gnu_winlibs = ['-lkernel32', '-luser32', '-lgdi32', '-lwinspool', '-lshell32',
@@ -248,6 +263,52 @@ clang_color_args = {'auto': ['-Xclang', '-fcolor-diagnostics'],
                     'never': ['-Xclang', '-fno-color-diagnostics'],
                     }
 
+arm_optimization_args = {'0': ['-O0'],
+                         'g': ['-g'],
+                         '1': ['-O1'],
+                         '2': ['-O2'],
+                         '3': ['-O3'],
+                         's': [],
+                         }
+
+armclang_optimization_args = {'0': ['-O0'],
+                              'g': ['-g'],
+                              '1': ['-O1'],
+                              '2': ['-O2'],
+                              '3': ['-O3'],
+                              's': ['-Os']
+                              }
+
+clike_optimization_args = {'0': [],
+                           'g': [],
+                           '1': ['-O1'],
+                           '2': ['-O2'],
+                           '3': ['-O3'],
+                           's': ['-Os'],
+                           }
+
+gnu_optimization_args = {'0': [],
+                         'g': ['-Og'],
+                         '1': ['-O1'],
+                         '2': ['-O2'],
+                         '3': ['-O3'],
+                         's': ['-Os'],
+                         }
+
+msvc_optimization_args = {'0': [],
+                          'g': ['/O0'],
+                          '1': ['/O1'],
+                          '2': ['/O2'],
+                          '3': ['/O2'],
+                          's': ['/O1'], # Implies /Os.
+                          }
+
+clike_debug_args = {False: [],
+                    True: ['-g']}
+
+msvc_debug_args = {False: [],
+                   True: []} # Fixme!
+
 base_options = {'b_pch': coredata.UserBooleanOption('b_pch', 'Use precompiled headers', True),
                 'b_lto': coredata.UserBooleanOption('b_lto', 'Use link time optimization', False),
                 'b_sanitize': coredata.UserComboOption('b_sanitize',
@@ -273,6 +334,9 @@ base_options = {'b_pch': coredata.UserBooleanOption('b_pch', 'Use precompiled he
                 'b_bitcode': coredata.UserBooleanOption('b_bitcode',
                                                         'Generate and embed bitcode (only macOS and iOS)',
                                                         False),
+                'b_vscrt': coredata.UserComboOption('b_vscrt', 'VS run-time library type to use.',
+                                                    ['none', 'md', 'mdd', 'mt', 'mtd', 'from_buildtype'],
+                                                    'from_buildtype'),
                 }
 
 gnulike_instruction_set_args = {'mmx': ['-mmmx'],
@@ -311,6 +375,14 @@ vs64_instruction_set_args = {'mmx': ['/arch:AVX'],
                              'neon': None,
                              }
 
+gnu_symbol_visibility_args = {'': [],
+                              'default': ['-fvisibility=default'],
+                              'internal': ['-fvisibility=internal'],
+                              'hidden': ['-fvisibility=hidden'],
+                              'protected': ['-fvisibility=protected'],
+                              'inlineshidden': ['-fvisibility=hidden', '-fvisibility-inlines-hidden'],
+                              }
+
 def sanitizer_compile_args(value):
     if value == 'none':
         return []
@@ -372,6 +444,15 @@ def get_base_compile_args(options, compiler):
     # This does not need a try...except
     if option_enabled(compiler.base_options, options, 'b_bitcode'):
         args.append('-fembed-bitcode')
+    try:
+        crt_val = options['b_vscrt'].value
+        buildtype = options['buildtype'].value
+        try:
+            args += compiler.get_crt_compile_args(crt_val, buildtype)
+        except AttributeError:
+            pass
+    except KeyError:
+        pass
     return args
 
 def get_base_link_args(options, linker, is_shared_module):
@@ -411,6 +492,15 @@ def get_base_link_args(options, linker, is_shared_module):
     elif as_needed:
         # -Wl,-dead_strip_dylibs is incompatible with bitcode
         args.append(linker.get_asneeded_args())
+    try:
+        crt_val = options['b_vscrt'].value
+        buildtype = options['buildtype'].value
+        try:
+            args += linker.get_crt_link_args(crt_val, buildtype)
+        except AttributeError:
+            pass
+    except KeyError:
+        pass
     return args
 
 class CrossNoRunException(MesonException):
@@ -464,7 +554,7 @@ class CompilerArgs(list):
     dedup2_suffixes = ()
     dedup2_args = ()
     # Arg prefixes and args that must be de-duped by returning 1
-    dedup1_prefixes = ('-l',)
+    dedup1_prefixes = ('-l', '-Wl,-l')
     dedup1_suffixes = ('.lib', '.dll', '.so', '.dylib', '.a')
     # Match a .so of the form path/to/libfoo.so.0.1.0
     # Only UNIX shared libraries require this. Others have a fixed extension.
@@ -568,7 +658,7 @@ class CompilerArgs(list):
             group_start = -1
             group_end = -1
             for i, each in enumerate(new):
-                if not each.startswith('-l') and not each.endswith('.a') and \
+                if not each.startswith(('-Wl,-l', '-l')) and not each.endswith('.a') and \
                    not soregex.match(each):
                     continue
                 group_end = i
@@ -866,7 +956,7 @@ class Compiler:
         raise EnvironmentException('Language {} does not support library finding.'.format(self.get_display_language()))
 
     def get_library_dirs(self, *args, **kwargs):
-        return []
+        return ()
 
     def has_multi_arguments(self, args, env):
         raise EnvironmentException(
@@ -999,15 +1089,17 @@ class Compiler:
         return None
 
     def build_osx_rpath_args(self, build_dir, rpath_paths, build_rpath):
+        # Ensure that there is enough space for large RPATHs and install_name
+        args = ['-Wl,-headerpad_max_install_names']
         if not rpath_paths and not build_rpath:
-            return []
+            return args
         # On OSX, rpaths must be absolute.
         abs_rpaths = [os.path.join(build_dir, p) for p in rpath_paths]
         if build_rpath != '':
             abs_rpaths.append(build_rpath)
-        # Ensure that there is enough space for large RPATHs
-        args = ['-Wl,-headerpad_max_install_names']
-        args += ['-Wl,-rpath,' + rp for rp in abs_rpaths]
+        # Need to deduplicate abs_rpaths, as rpath_paths and
+        # build_rpath are not guaranteed to be disjoint sets
+        args += ['-Wl,-rpath,' + rp for rp in OrderedSet(abs_rpaths)]
         return args
 
     def build_unix_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
@@ -1072,26 +1164,45 @@ class Compiler:
         # building fails with undefined symbols.
         return []
 
-GCC_STANDARD = 0
-GCC_OSX = 1
-GCC_MINGW = 2
-GCC_CYGWIN = 3
+    def gnu_symbol_visibility_args(self, vistype):
+        return []
 
-CLANG_STANDARD = 0
-CLANG_OSX = 1
-CLANG_WIN = 2
-# Possibly clang-cl?
+    def get_gui_app_args(self, value):
+        return []
 
-ICC_STANDARD = 0
-ICC_OSX = 1
-ICC_WIN = 2
+    def has_func_attribute(self, name, env):
+        raise EnvironmentException(
+            'Language {} does not support function attributes.'.format(self.get_display_language()))
+
+
+@enum.unique
+class CompilerType(enum.Enum):
+    GCC_STANDARD = 0
+    GCC_OSX = 1
+    GCC_MINGW = 2
+    GCC_CYGWIN = 3
+
+    CLANG_STANDARD = 10
+    CLANG_OSX = 11
+    CLANG_MINGW = 12
+    # Possibly clang-cl?
+
+    ICC_STANDARD = 20
+    ICC_OSX = 21
+    ICC_WIN = 22
+
+    @property
+    def is_standard_compiler(self):
+        return self.name in ('GCC_STANDARD', 'CLANG_STANDARD', 'ICC_STANDARD')
+
+    @property
+    def is_osx_compiler(self):
+        return self.name in ('GCC_OSX', 'CLANG_OSX', 'ICC_OSX')
+
+    @property
+    def is_windows_compiler(self):
+        return self.name in ('GCC_MINGW', 'GCC_CYGWIN', 'CLANG_MINGW', 'ICC_WIN')
 
-# GNU ld cannot be installed on macOS
-# https://github.com/Homebrew/homebrew-core/issues/17794#issuecomment-328174395
-# Hence, we don't need to differentiate between OS and ld
-# for the sake of adding as-needed support
-GNU_LD_AS_NEEDED = '-Wl,--as-needed'
-APPLE_LD_AS_NEEDED = '-Wl,-dead_strip_dylibs'
 
 def get_macos_dylib_install_name(prefix, shlib_name, suffix, soversion):
     install_name = prefix + shlib_name
@@ -1100,39 +1211,40 @@ def get_macos_dylib_install_name(prefix, shlib_name, suffix, soversion):
     install_name += '.dylib'
     return '@rpath/' + install_name
 
-def get_gcc_soname_args(gcc_type, prefix, shlib_name, suffix, soversion, is_shared_module):
-    if soversion is None:
-        sostr = ''
-    else:
-        sostr = '.' + soversion
-    if gcc_type == GCC_STANDARD:
+def get_gcc_soname_args(compiler_type, prefix, shlib_name, suffix, soversion, darwin_versions, is_shared_module):
+    if compiler_type.is_standard_compiler:
+        sostr = '' if soversion is None else '.' + soversion
         return ['-Wl,-soname,%s%s.%s%s' % (prefix, shlib_name, suffix, sostr)]
-    elif gcc_type in (GCC_MINGW, GCC_CYGWIN):
+    elif compiler_type.is_windows_compiler:
         # For PE/COFF the soname argument has no effect with GNU LD
         return []
-    elif gcc_type == GCC_OSX:
+    elif compiler_type.is_osx_compiler:
         if is_shared_module:
             return []
         name = get_macos_dylib_install_name(prefix, shlib_name, suffix, soversion)
-        return ['-install_name', name]
+        args = ['-install_name', name]
+        if darwin_versions:
+            args += ['-compatibility_version', darwin_versions[0], '-current_version', darwin_versions[1]]
+        return args
     else:
         raise RuntimeError('Not implemented yet.')
 
 def get_compiler_is_linuxlike(compiler):
-    if (getattr(compiler, 'gcc_type', None) == GCC_STANDARD) or \
-       (getattr(compiler, 'clang_type', None) == CLANG_STANDARD) or \
-       (getattr(compiler, 'icc_type', None) == ICC_STANDARD):
-        return True
-    return False
+    compiler_type = getattr(compiler, 'compiler_type', None)
+    return compiler_type and compiler_type.is_standard_compiler
 
 def get_compiler_uses_gnuld(c):
     # FIXME: Perhaps we should detect the linker in the environment?
     # FIXME: Assumes that *BSD use GNU ld, but they might start using lld soon
-    if (getattr(c, 'gcc_type', None) in (GCC_STANDARD, GCC_MINGW, GCC_CYGWIN)) or \
-       (getattr(c, 'clang_type', None) in (CLANG_STANDARD, CLANG_WIN)) or \
-       (getattr(c, 'icc_type', None) in (ICC_STANDARD, ICC_WIN)):
-        return True
-    return False
+    compiler_type = getattr(c, 'compiler_type', None)
+    return compiler_type in (
+        CompilerType.GCC_STANDARD,
+        CompilerType.GCC_MINGW,
+        CompilerType.GCC_CYGWIN,
+        CompilerType.CLANG_STANDARD,
+        CompilerType.CLANG_MINGW,
+        CompilerType.ICC_STANDARD,
+        CompilerType.ICC_WIN)
 
 def get_largefile_args(compiler):
     '''
@@ -1191,29 +1303,103 @@ def gnulike_default_include_dirs(compiler, lang):
         mlog.warning('No include directory found parsing "{cmd}" output'.format(cmd=" ".join(cmd)))
     return paths
 
-class GnuCompiler:
-    # Functionality that is common to all GNU family compilers.
-    def __init__(self, gcc_type, defines):
-        self.id = 'gcc'
-        self.gcc_type = gcc_type
-        self.defines = defines or {}
+
+class GnuLikeCompiler(abc.ABC):
+    """
+    GnuLikeCompiler is a common interface to all compilers implementing
+    the GNU-style commandline interface. This includes GCC, Clang
+    and ICC. Certain functionality between them is different and requires
+    that the actual concrete subclass define their own implementation.
+    """
+    def __init__(self, compiler_type):
+        self.compiler_type = compiler_type
         self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage',
-                             'b_colorout', 'b_ndebug', 'b_staticpic']
-        if self.gcc_type == GCC_OSX:
-            self.base_options.append('b_bitcode')
-        else:
+                             'b_ndebug', 'b_staticpic', 'b_asneeded']
+        if not self.compiler_type.is_osx_compiler:
             self.base_options.append('b_lundef')
-        self.base_options.append('b_asneeded')
-        # All GCC backends can do assembly
+        # All GCC-like backends can do assembly
         self.can_compile_suffixes.add('s')
 
-    # TODO: centralise this policy more globally, instead
-    # of fragmenting it into GnuCompiler and ClangCompiler
     def get_asneeded_args(self):
-        if self.gcc_type == GCC_OSX:
-            return APPLE_LD_AS_NEEDED
+        # GNU ld cannot be installed on macOS
+        # https://github.com/Homebrew/homebrew-core/issues/17794#issuecomment-328174395
+        # Hence, we don't need to differentiate between OS and ld
+        # for the sake of adding as-needed support
+        if self.compiler_type.is_osx_compiler:
+            return '-Wl,-dead_strip_dylibs'
         else:
-            return GNU_LD_AS_NEEDED
+            return '-Wl,--as-needed'
+
+    def get_pic_args(self):
+        if self.compiler_type.is_osx_compiler or self.compiler_type.is_windows_compiler:
+            return [] # On Window and OS X, pic is always on.
+        return ['-fPIC']
+
+    def get_buildtype_args(self, buildtype):
+        return gnulike_buildtype_args[buildtype]
+
+    @abc.abstractmethod
+    def get_optimization_args(self, optimization_level):
+        raise NotImplementedError("get_optimization_args not implemented")
+
+    def get_debug_args(self, is_debug):
+        return clike_debug_args[is_debug]
+
+    def get_buildtype_linker_args(self, buildtype):
+        if self.compiler_type.is_osx_compiler:
+            return apple_buildtype_linker_args[buildtype]
+        return gnulike_buildtype_linker_args[buildtype]
+
+    @abc.abstractmethod
+    def get_pch_suffix(self):
+        raise NotImplementedError("get_pch_suffix not implemented")
+
+    def split_shlib_to_parts(self, fname):
+        return os.path.dirname(fname), fname
+
+    def get_soname_args(self, *args):
+        return get_gcc_soname_args(self.compiler_type, *args)
+
+    def get_std_shared_lib_link_args(self):
+        return ['-shared']
+
+    def get_std_shared_module_link_args(self, options):
+        if self.compiler_type.is_osx_compiler:
+            return ['-bundle', '-Wl,-undefined,dynamic_lookup']
+        return ['-shared']
+
+    def get_link_whole_for(self, args):
+        if self.compiler_type.is_osx_compiler:
+            result = []
+            for a in args:
+                result += ['-Wl,-force_load', a]
+            return result
+        return ['-Wl,--whole-archive'] + args + ['-Wl,--no-whole-archive']
+
+    def get_instruction_set_args(self, instruction_set):
+        return gnulike_instruction_set_args.get(instruction_set, None)
+
+    def get_default_include_dirs(self):
+        return gnulike_default_include_dirs(self.exelist, self.language)
+
+    @abc.abstractmethod
+    def openmp_flags(self):
+        raise NotImplementedError("openmp_flags not implemented")
+
+    def gnu_symbol_visibility_args(self, vistype):
+        return gnu_symbol_visibility_args[vistype]
+
+
+class GnuCompiler(GnuLikeCompiler):
+    """
+    GnuCompiler represents an actual GCC in its many incarnations.
+    Compilers imitating GCC (Clang/Intel) should use the GnuLikeCompiler ABC.
+    """
+    def __init__(self, compiler_type, defines):
+        super().__init__(compiler_type)
+        self.id = 'gcc'
+        self.defines = defines or {}
+        self.base_options.append('b_colorout')
 
     def get_colorout_args(self, colortype):
         if mesonlib.version_compare(self.version, '>=4.9.0'):
@@ -1235,55 +1421,27 @@ class GnuCompiler:
         if define in self.defines:
             return self.defines[define]
 
-    def get_pic_args(self):
-        if self.gcc_type in (GCC_CYGWIN, GCC_MINGW, GCC_OSX):
-            return [] # On Window and OS X, pic is always on.
-        return ['-fPIC']
-
-    def get_buildtype_args(self, buildtype):
-        return gnulike_buildtype_args[buildtype]
-
-    def get_buildtype_linker_args(self, buildtype):
-        if self.gcc_type == GCC_OSX:
-            return apple_buildtype_linker_args[buildtype]
-        return gnulike_buildtype_linker_args[buildtype]
+    def get_optimization_args(self, optimization_level):
+        return gnu_optimization_args[optimization_level]
 
     def get_pch_suffix(self):
         return 'gch'
 
-    def split_shlib_to_parts(self, fname):
-        return os.path.dirname(fname), fname
-
-    def get_soname_args(self, prefix, shlib_name, suffix, soversion, is_shared_module):
-        return get_gcc_soname_args(self.gcc_type, prefix, shlib_name, suffix, soversion, is_shared_module)
-
-    def get_std_shared_lib_link_args(self):
-        return ['-shared']
-
-    def get_link_whole_for(self, args):
-        return ['-Wl,--whole-archive'] + args + ['-Wl,--no-whole-archive']
-
     def gen_vs_module_defs_args(self, defsfile):
         if not isinstance(defsfile, str):
             raise RuntimeError('Module definitions file should be str')
         # On Windows targets, .def files may be specified on the linker command
         # line like an object file.
-        if self.gcc_type in (GCC_CYGWIN, GCC_MINGW):
+        if self.compiler_type in (CompilerType.GCC_CYGWIN, CompilerType.GCC_MINGW):
             return [defsfile]
         # For other targets, discard the .def file.
         return []
 
-    def get_gui_app_args(self):
-        if self.gcc_type in (GCC_CYGWIN, GCC_MINGW):
+    def get_gui_app_args(self, value):
+        if self.compiler_type in (CompilerType.GCC_CYGWIN, CompilerType.GCC_MINGW) and value:
             return ['-mwindows']
         return []
 
-    def get_instruction_set_args(self, instruction_set):
-        return gnulike_instruction_set_args.get(instruction_set, None)
-
-    def get_default_include_dirs(self):
-        return gnulike_default_include_dirs(self.exelist, self.language)
-
     def openmp_flags(self):
         return ['-fopenmp']
 
@@ -1291,8 +1449,8 @@ class GnuCompiler:
 class ElbrusCompiler(GnuCompiler):
     # Elbrus compiler is nearly like GCC, but does not support
     # PCH, LTO, sanitizers and color output as of version 1.21.x.
-    def __init__(self, gcc_type, defines):
-        GnuCompiler.__init__(self, gcc_type, defines)
+    def __init__(self, compiler_type, defines):
+        GnuCompiler.__init__(self, compiler_type, defines)
         self.id = 'lcc'
         self.base_options = ['b_pgo', 'b_coverage',
                              'b_ndebug', 'b_staticpic',
@@ -1304,12 +1462,12 @@ class ElbrusCompiler(GnuCompiler):
         os_env = os.environ.copy()
         os_env['LC_ALL'] = 'C'
         stdo = Popen_safe(self.exelist + ['--print-search-dirs'], env=os_env)[1]
-        paths = []
+        paths = ()
         for line in stdo.split('\n'):
             if line.startswith('libraries:'):
                 # lcc does not include '=' in --print-search-dirs output.
                 libstr = line.split(' ', 1)[1]
-                paths = [os.path.realpath(p) for p in libstr.split(':')]
+                paths = (os.path.realpath(p) for p in libstr.split(':'))
                 break
         return paths
 
@@ -1317,52 +1475,31 @@ class ElbrusCompiler(GnuCompiler):
         os_env = os.environ.copy()
         os_env['LC_ALL'] = 'C'
         stdo = Popen_safe(self.exelist + ['--print-search-dirs'], env=os_env)[1]
-        paths = []
+        paths = ()
         for line in stdo.split('\n'):
             if line.startswith('programs:'):
                 # lcc does not include '=' in --print-search-dirs output.
                 libstr = line.split(' ', 1)[1]
-                paths = [os.path.realpath(p) for p in libstr.split(':')]
+                paths = (os.path.realpath(p) for p in libstr.split(':'))
                 break
         return paths
 
-class ClangCompiler:
-    def __init__(self, clang_type):
+
+class ClangCompiler(GnuLikeCompiler):
+    def __init__(self, compiler_type):
+        super().__init__(compiler_type)
         self.id = 'clang'
-        self.clang_type = clang_type
-        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage',
-                             'b_ndebug', 'b_staticpic', 'b_colorout']
-        if self.clang_type == CLANG_OSX:
+        self.base_options.append('b_colorout')
+        if self.compiler_type.is_osx_compiler:
             self.base_options.append('b_bitcode')
-        else:
-            self.base_options.append('b_lundef')
-        self.base_options.append('b_asneeded')
-        # All Clang backends can do assembly and LLVM IR
-        self.can_compile_suffixes.update(['ll', 's'])
-
-    # TODO: centralise this policy more globally, instead
-    # of fragmenting it into GnuCompiler and ClangCompiler
-    def get_asneeded_args(self):
-        if self.clang_type == CLANG_OSX:
-            return APPLE_LD_AS_NEEDED
-        else:
-            return GNU_LD_AS_NEEDED
-
-    def get_pic_args(self):
-        if self.clang_type in (CLANG_WIN, CLANG_OSX):
-            return [] # On Window and OS X, pic is always on.
-        return ['-fPIC']
+        # All Clang backends can also do LLVM IR
+        self.can_compile_suffixes.add('ll')
 
     def get_colorout_args(self, colortype):
         return clang_color_args[colortype][:]
 
-    def get_buildtype_args(self, buildtype):
-        return gnulike_buildtype_args[buildtype]
-
-    def get_buildtype_linker_args(self, buildtype):
-        if self.clang_type == CLANG_OSX:
-            return apple_buildtype_linker_args[buildtype]
-        return gnulike_buildtype_linker_args[buildtype]
+    def get_optimization_args(self, optimization_level):
+        return clike_optimization_args[optimization_level]
 
     def get_pch_suffix(self):
         return 'pch'
@@ -1373,17 +1510,6 @@ class ClangCompiler:
         # so it might change semantics at any time.
         return ['-include-pch', os.path.join(pch_dir, self.get_pch_name(header))]
 
-    def get_soname_args(self, prefix, shlib_name, suffix, soversion, is_shared_module):
-        if self.clang_type == CLANG_STANDARD:
-            gcc_type = GCC_STANDARD
-        elif self.clang_type == CLANG_OSX:
-            gcc_type = GCC_OSX
-        elif self.clang_type == CLANG_WIN:
-            gcc_type = GCC_MINGW
-        else:
-            raise MesonException('Unreachable code when converting clang type to gcc type.')
-        return get_gcc_soname_args(gcc_type, prefix, shlib_name, suffix, soversion, is_shared_module)
-
     def has_multi_arguments(self, args, env):
         myargs = ['-Werror=unknown-warning-option', '-Werror=unused-command-line-argument']
         if mesonlib.version_compare(self.version, '>=3.6.0'):
@@ -1399,29 +1525,10 @@ class ClangCompiler:
         # visibility to obey OS X and iOS minimum version targets with
         # -mmacosx-version-min, -miphoneos-version-min, etc.
         # https://github.com/Homebrew/homebrew-core/issues/3727
-        if self.clang_type == CLANG_OSX and version_compare(self.version, '>=8.0'):
+        if self.compiler_type.is_osx_compiler and version_compare(self.version, '>=8.0'):
             extra_args.append('-Wl,-no_weak_imports')
         return super().has_function(funcname, prefix, env, extra_args, dependencies)
 
-    def get_std_shared_module_link_args(self, options):
-        if self.clang_type == CLANG_OSX:
-            return ['-bundle', '-Wl,-undefined,dynamic_lookup']
-        return ['-shared']
-
-    def get_link_whole_for(self, args):
-        if self.clang_type == CLANG_OSX:
-            result = []
-            for a in args:
-                result += ['-Wl,-force_load', a]
-            return result
-        return ['-Wl,--whole-archive'] + args + ['-Wl,--no-whole-archive']
-
-    def get_instruction_set_args(self, instruction_set):
-        return gnulike_instruction_set_args.get(instruction_set, None)
-
-    def get_default_include_dirs(self):
-        return gnulike_default_include_dirs(self.exelist, self.language)
-
     def openmp_flags(self):
         if version_compare(self.version, '>=3.8.0'):
             return ['-fopenmp']
@@ -1505,26 +1612,22 @@ class ArmclangCompiler:
     def get_linker_exelist(self):
         return [self.linker_exe]
 
+    def get_optimization_args(self, optimization_level):
+        return armclang_optimization_args[optimization_level]
+
+    def get_debug_args(self, is_debug):
+        return clike_debug_args[is_debug]
+
 
 # Tested on linux for ICC 14.0.3, 15.0.6, 16.0.4, 17.0.1
-class IntelCompiler:
-    def __init__(self, icc_type):
+class IntelCompiler(GnuLikeCompiler):
+    def __init__(self, compiler_type):
+        super().__init__(compiler_type)
         self.id = 'intel'
-        self.icc_type = icc_type
         self.lang_header = 'none'
-        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage',
-                             'b_colorout', 'b_ndebug', 'b_staticpic', 'b_lundef', 'b_asneeded']
-        # Assembly
-        self.can_compile_suffixes.add('s')
 
-    def get_pic_args(self):
-        return ['-fPIC']
-
-    def get_buildtype_args(self, buildtype):
-        return gnulike_buildtype_args[buildtype]
-
-    def get_buildtype_linker_args(self, buildtype):
-        return gnulike_buildtype_linker_args[buildtype]
+    def get_optimization_args(self, optimization_level):
+        return clike_optimization_args[optimization_level]
 
     def get_pch_suffix(self):
         return 'pchi'
@@ -1536,45 +1639,15 @@ class IntelCompiler:
     def get_pch_name(self, header_name):
         return os.path.basename(header_name) + '.' + self.get_pch_suffix()
 
-    def split_shlib_to_parts(self, fname):
-        return os.path.dirname(fname), fname
-
-    def get_soname_args(self, prefix, shlib_name, suffix, soversion, is_shared_module):
-        if self.icc_type == ICC_STANDARD:
-            gcc_type = GCC_STANDARD
-        elif self.icc_type == ICC_OSX:
-            gcc_type = GCC_OSX
-        elif self.icc_type == ICC_WIN:
-            gcc_type = GCC_MINGW
-        else:
-            raise MesonException('Unreachable code when converting icc type to gcc type.')
-        return get_gcc_soname_args(gcc_type, prefix, shlib_name, suffix, soversion, is_shared_module)
-
-    # TODO: centralise this policy more globally, instead
-    # of fragmenting it into GnuCompiler and ClangCompiler
-    def get_asneeded_args(self):
-        if self.icc_type == CLANG_OSX:
-            return APPLE_LD_AS_NEEDED
-        else:
-            return GNU_LD_AS_NEEDED
-
-    def get_std_shared_lib_link_args(self):
-        # FIXME: Don't know how icc works on OSX
-        # if self.icc_type == ICC_OSX:
-        #     return ['-bundle']
-        return ['-shared']
-
-    def get_default_include_dirs(self):
-        return gnulike_default_include_dirs(self.exelist, self.language)
-
     def openmp_flags(self):
         if version_compare(self.version, '>=15.0.0'):
             return ['-qopenmp']
         else:
             return ['-openmp']
 
-    def get_link_whole_for(self, args):
-        return GnuCompiler.get_link_whole_for(self, args)
+    def has_arguments(self, args, env, code, mode):
+        # -diag-error 10148 is required to catch invalid -W options
+        return super().has_arguments(args + ['-diag-error', '10006', '-diag-error', '10148'], env, code, mode)
 
 
 class ArmCompiler:
@@ -1645,3 +1718,9 @@ class ArmCompiler:
 
     def get_coverage_link_args(self):
         return []
+
+    def get_optimization_args(self, optimization_level):
+        return arm_optimization_args[optimization_level]
+
+    def get_debug_args(self, is_debug):
+        return clike_debug_args[is_debug]
